#!python
#
# Usage:
#  python azalt_rate_time_tle.py <tle_filename>
#
# Uses ephem,      % pip install ephem
#      numpy,      % pip install numpy
#      spacetrack, % pip install spacetrack
#
# Observatory location:
# ---------------------
obs_lon = '-110.7889'
obs_lat = '32.44257'
obs_elev = 2805
#
# Read a two line element and return Az, Alt at a time,
# and compute differences to get rates
# and tracking bias.
# Original printed in format for input to CSS 40" scheduler, this
# version is more general and prints a table.  The table format is
# JD, UT hms, az (deg), alt, az rate in arcsec/sec, alt rate in arcsec/sec
# like:
# 2457672.6008 2016/10/11 02:25:11 20:37:34.84 -5:22:20.8    14.97    -0.02
#
# Alternate usage (see below, returns single az/alt):
#  python sat_altaz.py <tle_filename>  <+time_offset> <output_file>
#
# If tle_filename does not exist, fetch the TLE info from the Space-Track
# service using its REST API (and the spacetrack Python module :-))
# Save the just fetched TLE data into a local file, so the next time
# it's locally available. This creates a cache of TLE info here.
# PS the new,y created files will go into the directory that was 
# supplied in the 'fname' argument.
#
# Spacetrack Credentials:
# ----------------------
st_username = 'xxxxxxxx'
st_password = 'yyyyyyyy'
# ----------------------
#
# If the time_offset argument (format +nn) is given, the Az and Alt (only) are
# printed for a time 'nn' seconds from now. This mode is for use with real-time
# observing, where the 'nn' seconds provides a lead time for the Az/Alt slew
# so the target will be centered. The real-time observing engine will spawn
# this script when it adjusts the scope pointing immediately before acquiring
# each image. Obviously this could be extended to provide rates etc.
#
# If the output_file argument is given (requires time_offset argument), then
# the Az and Alt are printed to the given file. The file will be overwritten
# each time. 
#

#
# The TLE is read from a filename given on the command line; if not
# specified, user is prompted for a filename.
#
# Ben Weiner  Aug 2016
# Bob Denny   Feb 2017 for Schulman 32" auto-acquisition with live 
#             position updating. Add TLE cache and auto-fetch from
#             Space-Track. See expanded usage info above. 
#             I can be contacted at rdenny@dc3.com or +1 480 396 9700
# Bob Denny   Jan 2018 Remove canned lat/long and add user-edit above
#             Add more comments on the TLE cache handling etc.

import os
import sys
import string
import numpy as np
import ephem
from spacetrack import SpaceTrackClient

# Read lines from a file and hope that the first three are a TLE.
# Check for blank lines and comments with '#'
# SEE COMMENTS ABOVE FOR MISSING FILE AND TLE CACHE AND ALSO THOSE BELOW
#
def read_tle(fname):
    tlines = []
    # The fname argument is the full path to the TLE file in the TLECache folder, e.g.
    #     "C:\Users\whoever\Documents\ACP Astronomy\SatObs\TLECache\27715.tle"
    # If not there, chop the path up to get the number and use that to get the TLE (3LE)
    # from Spacetrack then make a new TLE file for the cache. Otherwise, read the TLE
    # from the cached file.
    #
    if not os.path.exists(fname):
        noext, ext = os.path.splitext(fname)
        cache_dir, sat_num = os.path.split(noext)
        #
        # http://spacetrack.readthedocs.io/en/latest/
        # https://pypi.python.org/pypi/spacetrack/
        # https://www.space-track.org/basicspacedata/query/class/tle_latest/ORDINAL/1/NORAD_CAT_ID/nnnnn/format/3le"
        #
        st = SpaceTrackClient(identity=st_username, password=st_password)
        tle_txt = st.tle_latest(ordinal=1, norad_cat_id=sat_num, format='3le').encode('ascii')
        f = open(fname, 'w')
        f.write(tle_txt)
        f.close()
        tlines = string.split(tle_txt, '\n')
    else:
        f = open(fname,'r')
        for line in f:
            if (line.strip() != '' and line[0] != '#'):
                tlines.append(line)
        f.close()
    tle = ephem.readtle(tlines[0], tlines[1], tlines[2])
    return tle

def get_a_time(localhms):
    tnow = ephem.now()
    if len(localhms) == 0:
        localhms = raw_input('Enter time to compute as UT: yyyy mm dd hh mm ss, or as time from now: +ss [now]: ')
    if len(localhms) == 19:             # Exactly as above time format
        localtime = localhms.split()
        # print localtime
        trequest_tuple = (int(localtime[0]), int(localtime[1]), int(localtime[2]), int(localtime[3]), int(localtime[4]), float(localtime[5]))
        trequest = ephem.Date(trequest_tuple)
    elif localhms.find('+') == 0:       # Starts with + means positive offset
        trequest = tnow + (string.atoi(localhms[1:]) * ephem.second)
        # print tnow, ephem.date(trequest)
    else:
        trequest = tnow
    return trequest

    
# From the TLE and observatory location, find ra, dec, and az, alt

# The coords being returned seemed correct in az, alt
# and off in ra, dec - turns out for satellites the ra, dec is in
# current epoch rather than J2000, see
# http://rhodesmill.org/pyephem/radec.html
# For satellites, I want apparent topocentric position, but pyephem
# always puts this in epoch-of-date, so need to precess it back to
# 2000 for entering into a TCS.

def compute_radec_azalt(tle, obs_location):
    radtodeg = 180.0 / ephem.pi
    tle.compute(obs_location)
    tnow = ephem.now()
    ra, dec = tle.ra, tle.dec
    ra2000, dec2000 = precess_now_2000(tle.ra, tle.dec)
    print 'NOW: RA, Dec = ', tle.ra, tle.dec
    print 'NOW: RA, Dec 2000 = ', ra2000, dec2000
    print 'NOW: az, alt = ', tle.az, tle.alt
    print 'NOW: az, alt =  %10.5f  %10.5f' % (float(tle.az)*radtodeg, float(tle.alt)*radtodeg)
#    t_plus_10min = ephem.now() + 10.0 / (24.0 * 60.0)
    t_plus_10min = ephem.now() + 10.0 * ephem.minute
    #obs_plus_10 = obs_location.copy()
    obs_plus_10 = obs_location
    obs_plus_10.date = t_plus_10min
    tle.compute(obs_plus_10)
    ra2000, dec2000 = precess_now_2000(tle.ra, tle.dec)
    print 'Plus 10 min: RA, Dec = ', tle.ra, tle.dec
    print 'Plus 10 min: RA, Dec 2000 = ', ra2000, dec2000
    print 'Plus 10 min: az, alt = ', tle.az, tle.alt
    print 'Plus 10 min: az, alt =  %10.5f  %10.5f' % (float(tle.az)*radtodeg, float(tle.alt)*radtodeg)
    localhms = raw_input('Enter UT hh mm ss time to compute [skip]: ')
    # If a string was entered, try splitting it on spaces, and if
    # that doesn't work, split on colons, but if it's not a valid time
    # weird things may happen.
    if len(localhms) > 1:
        localtime = localhms.split()
        if len(localtime) <= 1:
            localtime = localhms.split(':')
        tnow = ephem.now()
        # Extract the y/m/d from first 3 pieces of tuple
        tnowtuple = tnow.tuple()
        trequest_tuple = (tnowtuple[0], tnowtuple[1], tnowtuple[2], float(localtime[0]), float(localtime[1]), float(localtime[2]))
        # print trequest_tuple
        trequest = ephem.Date(trequest_tuple)
        obs_request = obs_location
        obs_request.date = trequest
        tle.compute(obs_request)
        ra2000, dec2000 = precess_now_2000(tle.ra, tle.dec)
        print 't requested: RA, Dec = ', tle.ra, tle.dec
        print 't requested: RA, Dec 2000 = ', ra2000, dec2000
        print 't requested: az, alt = ', tle.az, tle.alt
        azdeg = float(tle.az)*radtodeg
        altdeg = float(tle.alt)*radtodeg
        print 't requested: az, alt =  %10.5f  %10.5f' % (azdeg, altdeg)
        
    return ra, dec, tle.az, tle.alt

def compute_radec_azalt_at_time(tle, obs_location, time):
    radtodeg = 180.0 / ephem.pi
    obs_time = obs_location
    obs_time.date = time
    tle.compute(obs_time)
    # This would assume time is close enough to now that the precession
    # to 2000 is accurate
    # ra2000, dec2000 = precess_now_2000(tle.ra, tle.dec)
    # or can precess from the requested time
    ra2000, dec2000 = precess_date_2000(tle.ra, tle.dec, time)
    # print '%22s : RA, Dec 2000 = %s %s' % (tle.name, ra2000, dec2000)
    # azdeg = float(tle.az)*radtodeg
    # altdeg = float(tle.alt)*radtodeg
    # print '%22s : az, alt =  %10.5f  %10.5f' % (tle.name, float(tle.az)*radtodeg, float(tle.alt)*radtodeg)
    return ra2000, dec2000, tle.az, tle.alt

# Calculate ra,dec position at given time, and at dtime seconds later.
# Difference the positions to get the object motion in rate and pa,
# and the tracking rate biases.  Print in a generic format
# rather than CSS

def compute_radec_azalt_diff(tle, obs_location, time, dtime):
    time0 = time
    time1 = time + dtime * ephem.second
    ra0, dec0, az0, alt0 = compute_radec_azalt_at_time(tle, obs_location, time0)
    ra1, dec1, az1, alt1 = compute_radec_azalt_at_time(tle, obs_location, time1)
    # compute differences, rate, pa, bias rates
    # difference in radians?
    radiff = ra1 - ra0
    decdiff = dec1 - dec0
    # print "RA, Dec motion in radians: ", radiff, decdiff
    radtodeg = 180.0 / 3.1416
    radtosec = 180.0 * 3600.0/3.1416
    radiffsec = radiff * radtosec * np.cos(dec0)
    decdiffsec = decdiff * radtosec
    azdiff = az1 - az0
    altdiff = alt1 - alt0
    azdiffsec = azdiff * radtosec
    altdiffsec = altdiff * radtosec
    # CSS-specific stuff
    exptime = 5.0
    minsep = 30.0
    npasses = 4
    dithers = '{0.0 0.0} {0.0 10.0} {-7.0 -7.0} {7.0 -7.0}'
    satid = tle.catalog_number # tle numid.  Truncate to 5 digits
    satid5digit = int(satid % 1e5)
    #satname = tle.name
    epoch = 2000.0
    filtname = 'CLEAR'
    # pa should be 0=north, 90=east. sign may not be correct; check this
    # pa = np.arctan2(decdiff,radiff) * 180 / 3.1416
    pa = np.arctan2(radiff,decdiff) * 180 / 3.1416
    if pa < 0.0:
        pa = pa + 360.0
    rate = np.sqrt(radiffsec**2 + decdiffsec**2) / (dtime / 60.0)
    rabias = radiffsec / dtime
    decbias = decdiffsec / dtime
    azrate = azdiffsec / dtime
    altrate = altdiffsec / dtime
    #time0float = float(time0)
    # ephem times are in Dublin JD = JD - 2415020
    jd_time0 = 2415020 + time0
    # Pick whether you want to constrain the earliest allowable observation
    # aftertime = jd_time0
    aftertime = 0.0
    # to print stuff in CSS format.  Comma at end of line suppresses the
    # line break.
    # print '{I52 {SO Stellarview} {} scp://css@scopenet.as.arizona.edu:/data40 %d {%s} %12.4f %5.1f}' % (npasses, dithers, aftertime, minsep),
    # print '{U%05d %s %s %7.2f %5.1f %s' % (satid5digit, ra0, dec0, epoch, exptime, filtname),
    # print '%12.4f %7.2f %7.2f %7.2f %7.2f}' % (jd_time0, pa, rate, rabias, decbias)
    # to print a generic format, which is JD, UT hms, ra, dec, rabias in arcsec/sec, Dec bias in arcsec/sec
    # print '%12.4f %s %s %s %8.2f %8.2f' % (jd_time0, time0, ra0, dec0, rabias, decbias)
    # to print a generic az-alt format, which is JD, UT hms, az(deg), alt(deg), rabias in arcsec/sec, Dec bias in arcsec/sec
    print '%12.4f %s %9.4f %9.4f %8.2f %8.2f' % (jd_time0, time0, az0*radtodeg, alt0*radtodeg, azrate, altrate)
    # return ra0, dec0, pa, rate
    return az0, alt0, azrate, altrate

def precess_now_2000(ranow, decnow):
    coo = ephem.Equatorial(ranow, decnow, epoch=ephem.now())
    coo2 = ephem.Equatorial(coo, epoch='2000')
    return coo2.ra, coo2.dec

def precess_date_2000(radate, decdate, date):
    # date is an ephem.Date format
    coo = ephem.Equatorial(radate, decdate, epoch=date)
    coo2 = ephem.Equatorial(coo, epoch='2000')
    return coo2.ra, coo2.dec

def get_loop_params():
    looptimestring = raw_input('Enter total duration, interval to print, and look-ahead for rate, in min [30 2 0.5]: ')
    if len(looptimestring) > 1:
        looptimes = looptimestring.split()
        looplength = float(looptimes[0]) * 60.0
        loopstep   = float(looptimes[1]) * 60.0
        dtime      = float(looptimes[2]) * 60.0
    else:
        looplength = 1800.0
        loopstep = 120.0
        dtime = 30.0
    return looplength, loopstep, dtime

def print_rates_loop(tle, obs_location, timestart):
    # prompt for times to loop through
    looplength, loopstep, dtime = get_loop_params()
    nstep = int(looplength / loopstep)
    print ' JD time     date   UT hh:mm:ss  Az        Alt       Az_rate Alt_rate'
    for i in range(nstep):
        time1 = ephem.Date(timestart + i * loopstep * ephem.second)
        # ra, dec, pa, rate = compute_radec_diff(tle, obs_location, time1, dtime)
        az, alt, azrate, altrate = compute_radec_azalt_diff(tle, obs_location, time1, dtime)
    return

def print_single_altaz(tle, obs_location, time, ofile):
    radtodeg = 180.0 / 3.1416
    ra, dec, az, alt = compute_radec_azalt_at_time(tle, obs_location, time)
    ostring = '%1.4f,%1.4f' % (az * radtodeg, alt * radtodeg)   # Parseable as JS
    if len(ofile) > 0:
        f = open(ofile, 'w')
        f.write(ostring + '\n')
        f.close()
    else:
        print  ostring 
    return

def main():
    timearg = ""
    ofile = ""
    if len(sys.argv) >= 2:
        fname = sys.argv[1]
        if len(sys.argv) >= 3:
            timearg = sys.argv[2]
            if len(sys.argv) >= 4:
                ofile = sys.argv[3]
    else:
        fname = raw_input('Enter filename with TLE: ')
    tle = read_tle(fname)
    observatory = ephem.Observer()
    observatory.lat = obs_lat
    observatory.lon = obs_lon
    observatory.elevation = obs_elev
    time = get_a_time(timearg)
    if len(sys.argv) >= 3:
        print_single_altaz(tle, observatory, time, ofile)
    else:
        print_rates_loop(tle, observatory, time)
    return

# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
  main()
